# generated by datamodel-codegen:
#   filename:  csm_api-swagger.json

from __future__ import annotations

from enum import Enum
from typing import Dict, List, Optional

import cherwell_pydantic_api.types as ct
from cherwell_pydantic_api.generated_api_utils import ApiBaseModel, HttpStatusCode

from .... import CherwellObjectID
from . import Core


class ApprovalReadResponse(ApiBaseModel):
    properties: Optional[Dict[str, CherwellObjectID]] = None
    links: Optional[List[Core.CherwellLink]] = None
    errorCode: Optional[str] = None
    errorMessage: Optional[str] = None
    hasError: Optional[bool] = None
    httpStatusCode: Optional[HttpStatusCode] = None


import cherwell_pydantic_api.types as ct


class GetApprovalsResponse(ApiBaseModel):
    totalRecords: Optional[int] = None
    approvals: Optional[List[ApprovalReadResponse]] = None
    errorCode: Optional[str] = None
    errorMessage: Optional[str] = None
    hasError: Optional[bool] = None
    httpStatusCode: Optional[HttpStatusCode] = None
