# Auto-generated file. Do not edit!
###################################


from typing import ClassVar
from cherwell_pydantic_api.bo.model_base import BusinessObjectModelBase


class BusinessObjectModelInstanceBase(BusinessObjectModelBase):
    """Base class for business object models within the {{ settings.name }} instance.
    Generated by cherwell_pydantic_api."""

    _instance_name: ClassVar[str] = "{{ settings.name }}"
    _subpackage: ClassVar[str] = "{{ settings.get_repo_subpackage() }}"

